from enum import Enum
from pathlib import Path
from typing import Type, cast, Iterable

from collections_extended import RangeMap
from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel, Field
from pymultirole_plugins.v1.formatter import FormatterParameters, FormatterBase
from pymultirole_plugins.v1.schema import Document, Annotation
from starlette.responses import Response, JSONResponse


class ConsolidationType(str, Enum):
    linker = 'linker'
    unknown = 'unknown'
    unknown_only = 'unknown_only'


class ConsolidateParameters(FormatterParameters):
    type: ConsolidationType = Field(ConsolidationType.unknown, description="""Type of consolidation, use<br />
    - *linker* to retain only known entities<br />
    - *unknown* to retain all but prepending `unknown_` to the label name of unknwon entities<br />
    - *unknown_only* to retain only unknown entities prepending `unknown_` to their label name""")


class ConsolidateFormatter(FormatterBase):
    """Consolidate formatter .
    """

    def format(self, document: Document, parameters: FormatterParameters) \
            -> Response:
        def has_knowledge(a: Annotation):
            return a.terms is not None or a.properties is not None

        params: ConsolidateParameters = cast(ConsolidateParameters, parameters)
        if document.annotations:
            anns = self.filter_annotations(document)
            if params.type in [ConsolidationType.unknown_only, ConsolidationType.unknown]:
                for a in anns:
                    if not has_knowledge(a):
                        a.labelName = "unknown_" + a.labelName
                        if a.label:
                            a.label = "Unknown " + a.label
                if params.type == ConsolidationType.unknown_only:
                    anns = [a for a in anns if not has_knowledge(a)]
            elif params.type == ConsolidationType.linker:
                anns = [a for a in anns if has_knowledge(a)]
            document.annotations = anns
        if document.properties and "fileName" in document.properties:
            filepath = Path(document.properties['fileName'])
            filename = f"{filepath.stem}.json"
        elif document.identifier:
            filename = f"{document.identifier}.json"
        else:
            filename = "file.json"
        resp = JSONResponse(content=jsonable_encoder(document))
        resp.headers["Content-Disposition"] = f"attachment; filename={filename}"
        return resp

    @classmethod
    def get_model(cls) -> Type[BaseModel]:
        return ConsolidateParameters

    def filter_annotations(self, input: Document):
        """Filter a sequence of annotations and remove duplicates or overlaps. When spans overlap, the (first)
        longest span is preferred over shorter spans.
        annotations (iterable): The annotations to filter.
        RETURNS (list): The filtered annotations.
        """

        def get_sort_key(a: Annotation):
            return a.end - a.start, -a.start, a.labelName is None

        sorted_annotations: Iterable[Annotation] = sorted(input.annotations, key=get_sort_key, reverse=True)
        result = []
        seen_offsets = RangeMap()
        for ann in sorted_annotations:
            # Check for end - 1 here because boundaries are inclusive
            if seen_offsets.get(ann.start) is None and seen_offsets.get(ann.end - 1) is None:
                if ann.text is None:
                    ann.text = input.text[ann.start:ann.end]
                result.append(ann)
                seen_offsets[ann.start:ann.end] = ann
            else:
                target = seen_offsets.get(ann.start) or seen_offsets.get(ann.end - 1)
                # if target.labelName in kb_labels and ann.labelName in white_labels and (target.start-ann.start != 0 or target.end-ann.end != 0):
                if (target.start - ann.start == 0 or target.end - ann.end == 0) and (ann.end - ann.start) / (
                        target.end - target.start) > 0.8:
                    if ann.terms:
                        terms = set(target.terms or [])
                        terms.update(ann.terms)
                        target.terms = list(terms)
                    if ann.properties:
                        props = target.properties or {}
                        props.update(ann.properties)
                        target.properties = props
        result = sorted(result, key=lambda ann: ann.start)
        return result



import json
from collections import defaultdict
from itertools import groupby
from pathlib import Path

import pytest
from pymultirole_plugins.v1.schema import Document, Annotation
from starlette.responses import JSONResponse

from pyformatters_consolidate.consolidate import ConsolidateFormatter, ConsolidateParameters, ConsolidationType


def by_lexicon(a: Annotation):
    if a.terms:
        return a.terms[0].lexicon
    else:
        return ""


def by_label(a: Annotation):
    return a.labelName


def group_annotations(doc: Document, keyfunc):
    groups = defaultdict(list)
    for k, g in groupby(sorted(doc.annotations, key=keyfunc), keyfunc):
        groups[k] = list(g)
    return groups


def test_model():
    model = ConsolidateFormatter.get_model()
    model_class = model.construct().__class__
    assert model_class == ConsolidateParameters


# Arrange
@pytest.fixture
def original_doc():
    testdir = Path(__file__).parent
    source = Path(testdir, 'data/response_1622736571452.json')
    with source.open("r") as fin:
        docs = json.load(fin)
        original_doc = Document(**docs[0])
        return original_doc


def assert_original_doc(original_doc):
    original_groups_lexicon = group_annotations(original_doc, by_lexicon)
    original_groups_label = group_annotations(original_doc, by_label)
    assert len(original_groups_lexicon[""]) == 22
    assert len(original_groups_lexicon["lex_location"]) == 1
    assert len(original_groups_lexicon["lex_person"]) == 8
    assert len(original_groups_label["organization"]) == 4
    assert len(original_groups_label["location"]) == 4
    assert len(original_groups_label["person"]) == 23
    return original_groups_lexicon, original_groups_label


def test_consolidate_linker(original_doc):
    # linker
    original_groups_lexicon, original_groups_label = assert_original_doc(original_doc)
    assert original_groups_lexicon is not None
    assert original_groups_label is not None
    doc = original_doc.copy(deep=True)
    formatter = ConsolidateFormatter()
    parameters = ConsolidateParameters(type=ConsolidationType.linker)
    resp: JSONResponse = formatter.format(doc, parameters)
    content = resp.body.decode(resp.charset)
    consolidated: Document = Document(**json.loads(content))
    consolidated_groups_lexicon = group_annotations(consolidated, by_lexicon)
    # consolidated_groups_label = group_annotations(consolidated, by_label)
    assert resp.status_code == 200
    assert resp.media_type == "application/json"
    assert len(original_doc.annotations) > len(consolidated.annotations)
    assert len(consolidated_groups_lexicon[""]) == 0
    assert "file.json" in resp.headers['Content-Disposition']


def test_consolidate_unknown(original_doc):
    # unknown
    original_groups_lexicon, original_groups_label = assert_original_doc(original_doc)
    assert original_groups_lexicon is not None
    assert original_groups_label is not None
    doc = original_doc.copy(deep=True)
    formatter = ConsolidateFormatter()
    parameters = ConsolidateParameters(type=ConsolidationType.unknown)
    resp: JSONResponse = formatter.format(doc, parameters)
    content = resp.body.decode(resp.charset)
    consolidated: Document = Document(**json.loads(content))
    # consolidated_groups_lexicon = group_annotations(consolidated, by_lexicon)
    consolidated_groups_label = group_annotations(consolidated, by_label)
    assert resp.status_code == 200
    assert resp.media_type == "application/json"
    assert len(original_doc.annotations) > len(consolidated.annotations)
    knowns_unknowns = defaultdict(list)
    for label, anns in consolidated_groups_label.items():
        k = 'unknown' if label.startswith("unknown_") else "known"
        knowns_unknowns[k].extend(anns)
    assert len(knowns_unknowns["known"]) > 0
    assert len(knowns_unknowns["unknown"]) > 0


def test_consolidate_unknown_only(original_doc):
    # unknwon only
    original_groups_lexicon, original_groups_label = assert_original_doc(original_doc)
    assert original_groups_lexicon is not None
    assert original_groups_label is not None
    doc = original_doc.copy(deep=True)
    formatter = ConsolidateFormatter()
    parameters = ConsolidateParameters(type=ConsolidationType.unknown_only)
    resp: JSONResponse = formatter.format(doc, parameters)
    content = resp.body.decode(resp.charset)
    consolidated: Document = Document(**json.loads(content))
    # consolidated_groups_lexicon = group_annotations(consolidated, by_lexicon)
    consolidated_groups_label = group_annotations(consolidated, by_label)
    assert resp.status_code == 200
    assert resp.media_type == "application/json"
    assert len(original_doc.annotations) > len(consolidated.annotations)
    knowns_unknowns = defaultdict(list)
    for label, anns in consolidated_groups_label.items():
        k = 'unknown' if label.startswith("unknown_") else "known"
        knowns_unknowns[k].extend(anns)
    assert len(knowns_unknowns["known"]) == 0
    assert len(knowns_unknowns["unknown"]) > 0
